
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "globals.h"
#include "sym.h"

typedef struct constant {
	struct constant *next;
	void *data;
	int   size;
	int   uniq;
	char  seg;
} constant;

constant *constants;
symbol   *symbols;

int uniq;

int dataend;

dumpsyms(fp)
	FILE *fp;
{
	symbol *s;

	for (s = symbols; s; s = s->next) {
		if (s == symbols)
			fprintf(fp, "Symbols:\n");
		prsym(fp, s, NULL);
	}
}

int
storage(cc)
	int cc;
{
	int addr = dataend;

	dataend += cc;

	return (addr);
}

int
emits(str)
	char *str;
{
	symbol *s = addsym(NULL, 'S', strdup(str));

	return (s->uniq);
}

kill_scope()
{
	symbol *s, **sp;

	for (sp = &symbols; s = *sp; )
		switch (s->type) {
		case 'l': /* label */
		case 'p': /* parameter */
		case 'v': /* local variable */

			if (VERBOSE(4))
				prsym(stderr, s, "localkill");

			*sp = s->next;
			free(s->name);
			free(s);
			break;
		default:
			sp = &s->next;
		}
}

symbol *
symlookup(name)
	char *name;
{
	symbol *s;

	for (s = symbols; s; s = s->next)
		if (!strcmp(s->name, name))
			break;
	return (s);
}

symbol *
addsym(name, type, value)
	char *name;
	int type, value;
{
	symbol *s;

	if ((s = symlookup(name)) != NULL) {
		if (s->type != 0)
			error("%s already defined", name);
	} else {
		s = NEW(symbol);
		s->name = nullck(strdup(name));
		s->next = symbols;
		s->uniq = uniq++;
		symbols = s;
	}
	s->type  = type;
	s->value = value;

	if (VERBOSE(4))
		prsym(stderr, s, "addsym");

	return (s);
}

prsym(fp, s, add)
	FILE *fp;
	symbol *s;
	char *add;
{
	fprintf(fp, "%08x: %c %-20s %-12s\n",
		s->value, s->type ? s->type : '?', s->name,
		add ? add : "");
}

codeinit()
{
	dataend = 0;

	if (VERBOSE(5)) {
		extern int yydebug;
		yydebug = 1;
	}
	printf("_CODE\n");
}

codefini()
{
	symbol *s;
	constant *c;

	printf("_ROCONST\n");

	for (c = constants; c; c = c->next) {
			int i;
			u_char *p = c->data;

			printf("L%d: .ascii \"", c->uniq);
			for (i = 0; i < c->size; ++i)
				printf(isprint(p[i]) ? "%c" : "\\x%02X", p[i]);
			printf("\"\n");
	}
	printf("\n");

	printf("_DATA\n");

	if (VERBOSE(2))
		stdlist = stdout;

	if (stdlist)
		dumpsyms(stdlist);

	for (s = symbols; s; s = s->next) {
		if (s->type == 'V')
			if (s->size)
				printf("%s  VAR(%d)\n", s->name, s->size);
			else
				printf("%s  VAR\n", s->name);
	}
	printf("\n");
}

void
error(char *fmt, ...)
{
	va_list ap;
	FILE *fp = stderr;

	va_start(ap, fmt);

	fprintf(fp, "%s: ", __pname);
	vfprintf(fp, fmt, ap);
	fprintf(fp, "\n");

	exit(1);
}

static int func_end;
static int param_count;
static int local_count;

param_clear()
{
	param_count = local_count = 0;
}

var_ref(symbol *sym, int size)
{
	if (sym->type == 0) {
		sym->value = storage(sym->size);
		sym->type  = 'V';
	}
}

param_ref(char *name, int size)
{
	symbol *s;

	printf("%s = %d\n", name, param_count);
	s = addsym(name, 'p', param_count++);
	s->size = size;
	s->seg  = 's';
}

local_ref(char *name, int size)
{
	symbol *s;

	printf("%s = %d\n", name, param_count + local_count);
	s = addsym(name, 'v', param_count + local_count++);
	s->size = size;
	s->seg  = 's';
}

enter_function(symbol *sym)
{
	printf("%s:\n", sym->name);
	printf("ENTER_%d\n", local_count);

	func_end = uniq++;
}

return_jump()
{
	printf("JMP L%d # return\n", func_end);
}

leave_function(symbol *sym)
{
	printf("L%d:\n", func_end);
	printf("LEAVE_%d\n", local_count);
	printf("RETURN\n");
	printf("\n");
}

fcall(sym, argc)
	symbol *sym;
{
	printf("CALL(%s)\n", sym->name);
	printf("DISPOSE_%d\n", argc);
}

instack(symbol *sym)
{
	return (sym->seg == 's');
}

pure_string(char *str)
{
	constant *c = NEW(constant);

	c->next = constants;
	constants = c;

	c->data = str;
	c->size = strlen(str) + 1;

	return (c->uniq = uniq++);
}

op_gen(symbol *s, char *op)
{
	if (s->type == 'p')
		printf("%s_arg_%d # %s\n", op, s->value, s->name);
	else
	if (s->seg == 's')
		printf("%s_stk_%d # %s\n", op, s->value, s->name);
	else
		printf("%s(%s)\n", op, s->name);
}

op_load(symbol *sym)
{
	op_gen(sym, "LOAD");
}

op_store(symbol *sym)
{
	op_gen(sym, "STORE");
}

op_preinc(symbol *sym)
{
	op_gen(sym, "PREINC");
}

op_predec(symbol *sym)
{
	op_gen(sym, "PREDEC");
}

op_postinc(symbol *sym)
{
	op_gen(sym, "POSTINC");
}

op_postdec(symbol *sym)
{
	op_gen(sym, "POSTDEC");
}

